#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
 File name: mod_in_nse.py
 Application: The NewsLookout Web Scraping Application
 Date: 2020-01-11
 Copyright 2020, The NewsLookout Web Scraping Application, Sandeep Singh Sandhu, sandeep.sandhu@gmx.com
 Country: India
 Purpose: Plugin for the National Stock Exchange Data, India
 Save the daily BhavCopy file generated by the exchange at the end of the day


 Notice:
 This software is intended for demonstration and educational purposes only. This software is
 experimental and a work in progress. Under no circumstances should these files be used in
 relation to any critical system(s). Use of these files is at your own risk.

 Before using it for web scraping any website, always consult that website's terms of use.
 Do not use this software to fetch any data from any website that has forbidden use of web
 scraping or similar mechanisms, or violates its terms of use in any other way. The author is
 not liable for such kind of inappropriate use of this software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
 INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
 FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
 OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 DEALINGS IN THE SOFTWARE.

"""

##########

# import standard python libraries:
from datetime import datetime
import os
import logging

# import web retrieval and text processing python libraries:
# from bs4 import BeautifulSoup
# import lxml
import zipfile


# import this project's python libraries:
from base_plugin import basePlugin
from scraper_utils import getPreviousDaysDate
from data_structs import Types

##########

logger = logging.getLogger(__name__)


class mod_in_nse(basePlugin):
    """ Web Scraping plugin: mod_in_nse
    For: National Stock Exchange
    """

    minArticleLengthInChars = 10000
    pluginType = Types.MODULE_DATA_CONTENT  # implies data content fetcher

    mainURL = 'https://www1.nseindia.com/archives/equities/bhavcopy/pr/PR'
    mainURL_suffix = ".zip"

    masterData = {'mod_in_nse_EQUITY_L.csv': 'https://archives.nseindia.com/content/equities/EQUITY_L.csv',
                  'mod_in_nse_SME_EQUITY_L.csv': 'https://archives.nseindia.com/emerge/corporates/content/SME_EQUITY_L.csv',
                  'mod_in_nse_eq_etfseclist.csv': 'https://archives.nseindia.com/content/equities/eq_etfseclist.csv',
                  'mod_in_nse_PREF.csv': 'https://archives.nseindia.com/content/equities/PREF.csv',
                  'mod_in_nse_namechange.csv': 'https://archives.nseindia.com/content/equities/namechange.csv',
                  'mod_in_nse_symbolchange.csv': 'https://archives.nseindia.com/content/equities/symbolchange.csv',
                  'mod_in_nse_WARRANT.csv': 'https://archives.nseindia.com/content/equities/WARRANT.csv',
                  'mod_in_nse_DEBT.csv': 'https://archives.nseindia.com/content/equities/DEBT.csv',
                  'mod_in_nse_Redmn_Def_Web.csv': 'https://archives.nseindia.com/content/debt/Redmn_Def_Web.csv'
                  }

    pledgesURL1 = 'https://www.nseindia.com/companies-listing/corporate-filings-pledged-data'
    pledgesURL2 = 'https://www.nseindia.com/api/corporate-pledgedata?index=equities&csv=true'
    pledgesURL3 = 'https://www.nseindia.com/companies-listing/corporate-filings-regulation-7-promoter'

    listOfURLS = []
    urlUniqueRegexps = ["(^https://www1.nseindia.com/archives/equities/bhavcopy/pr/PR)([0-9]+)(.zip$)"]
    urlMatchPatterns = []
    uRLdata = dict()
    authorRegexps = []
    articleDateRegexps = dict()
    validURLStringsToCheck = []
    invalidURLSubStrings = []
    nonContentURLs = []
    nonContentStrings = []
    allowedDomains = ["www1.nseindia.com", "www.nseindia.com"]

    def __init__(self):
        """ Initialize the object
        """
        self.count_history_to_fetch = 10
        super().__init__()

    def getURLsListForDate(self, runDate):
        """ Retrieve the URLs List For a given Date
        """
        self.listOfURLS = []
        try:
            businessDate = getPreviousDaysDate(runDate)

            urlForDate = self.mainURL + businessDate.strftime("%d%m%y") + self.mainURL_suffix

            self.listOfURLS = [urlForDate]

            logger.info("Total count of valid articles to be retrieved = %s for business date: %s",
                        len(self.listOfURLS), businessDate.strftime("%Y-%m-%d"))

        except Exception as e:
            logger.error("Error trying to retrieve listing of URLs: %s", e)

    def extractLinksFromURLList(self, runDate, listOfURLS):
        """ Generate history URLs by filling the default date pattern of the data file.
        """
        additionalLinks = []
        startDate = getPreviousDaysDate(runDate)
        # begin with previous date
        businessDate = startDate

        for dayCount in range(self.count_history_to_fetch):
            # decrement dates one by one
            businessDate = getPreviousDaysDate(businessDate)
            urlForDate = self.mainURL + businessDate.strftime("%d%m%y") + self.mainURL_suffix
            additionalLinks.append(urlForDate)

        logger.info("Generated links for %s dates", len(additionalLinks))
        return(additionalLinks)

    def fetchDataFromURL(self, uRLtoFetch, WorkerID):
        """ Fetch data From given URL
        """
        fullPathName = ""
        dirPathName = ""
        sizeOfDataDownloaded = -1
        uncompressSize = 0
        publishDateStr = ""
        # output tuple structure: (uRL, len_raw_data, len_text, publish_date)
        resultVal = (uRLtoFetch, None, None, None)
        logger.debug("Fetching %s, Worker ID %s", uRLtoFetch.encode("ascii"), WorkerID)
        try:
            (publishDate, dataUniqueID) = self.extractUniqueIDFromURL(uRLtoFetch)
            rawData = self.networkHelper.fetchRawDataFromURL(uRLtoFetch, type(self).__name__)
            # write data to file:
            fileNameWithOutExt = self.makeUniqueFileName(dataUniqueID)
            publishDateStr = str(publishDate.strftime("%Y-%m-%d"))
            dirPathName = os.path.join(self.configData['data_dir'], publishDateStr)
            fullPathName = os.path.join(dirPathName, fileNameWithOutExt + ".zip")
            sizeOfDataDownloaded = len(rawData)
        except Exception as e:
            logger.error("Trying to fetch data from given URL: %s", e)

        if sizeOfDataDownloaded > self.minArticleLengthInChars:
            try:
                if os.path.isdir(dirPathName) is False:
                    # since dir does not exist, so try creating it:
                    os.mkdir(dirPathName)
            except Exception as theError:
                logger.error("Error creating data directory '%s', Exception was: %s", dirPathName, theError)
            try:
                with open(fullPathName, 'wb') as fp:
                    n = fp.write(rawData)
                    logger.debug("Wrote %s bytes to file: %s", n, fullPathName)
                    fp.close()
                # save master data:
                sizeOfDataDownloaded = self.fetchMasterData(uRLtoFetch, dirPathName, WorkerID, sizeOfDataDownloaded)
                # save pledges data:
                sizeOfDataDownloaded = sizeOfDataDownloaded + self.fetchPledgesData(dirPathName, publishDate)
                uncompressSize = self.parseFetchedData(str(publishDate.strftime("%Y%m%d")),
                                                       fullPathName,
                                                       dirPathName,
                                                       WorkerID)
            except Exception as theError:
                logger.error("Error writing downloaded data to zip file '%s': %s", fullPathName, theError)
            resultVal = (uRLtoFetch, sizeOfDataDownloaded, uncompressSize, publishDateStr)
        else:
            logger.info("Ignoring data zip file '%s' since its size %s is less than %s bytes",
                        fullPathName, len(rawData), self.minArticleLengthInChars)
        return(resultVal)

    def fetchPledgesData(self, dirPathName, publishDate):
        """ fetch Pledges Data"""
        n = 0
        try:
            publishDateStr = str(publishDate.strftime("%Y-%m-%d"))
            # payload = {}
            # httpHeaders = {}
            # self.networkHelper.postHTTPData(self.pledgesURL1, payload, postHeaders=httpHeaders)
            httpsResp = self.networkHelper.getHTTPData(self.pledgesURL2)
            pledgesPathName = os.path.join(dirPathName, 'nse_pledges_' + publishDateStr + '.csv')
            with open(pledgesPathName, 'wb') as fp:
                n = fp.write(httpsResp.content)
                logger.debug("Wrote %s bytes to pledges file: %s", n, pledgesPathName)
                fp.close()
        except Exception as e:
            logger.error("Error saving pledges data: %s", e)
        return(n)

    def fetchMasterData(self, uRLtoFetch, dirPathName, WorkerID, sizeOfDataDownloaded):
        """ fetch Master Data"""
        for masterFileName in self.masterData:
            try:
                masterRawData = self.networkHelper.fetchRawDataFromURL(self.masterData[masterFileName],
                                                                       type(self).__name__)
                sizeOfDataDownloaded = sizeOfDataDownloaded + len(masterRawData)
                fullMasterPathName = os.path.join(dirPathName, masterFileName)
                with open(fullMasterPathName, 'wb') as fp:
                    n = fp.write(masterRawData)
                    logger.debug("Wrote %s bytes to file: %s", n, fullMasterPathName)
                    fp.close()
            except Exception as e:
                logger.error("Error saving master data: %s", e)
        return(sizeOfDataDownloaded)

    def parseFetchedData(self, publishDateStr, zipFileName, dataDirForDate, WorkerID):
        """Parse the fetched Data
        """
        uncompressSize = 0
        try:
            zipDatafile = zipfile.ZipFile(zipFileName, mode='r')
            logger.debug("Expanding the fetched Zip archive, WorkerID = %s", WorkerID)
            for memberZipInfo in zipDatafile.infolist():
                memberFileName = memberZipInfo.filename
                if memberFileName.find('Readme.txt') < 0:
                    try:
                        if os.path.isfile(os.path.join(dataDirForDate, memberFileName)) is False:
                            logger.debug("Extracting file '%s' from zip archive.", memberFileName)
                            zipDatafile.extract(memberZipInfo, path=dataDirForDate)
                        newFileName = self.calculateFileNamePrefix(memberFileName, dataDirForDate, publishDateStr)
                        if os.path.isfile(newFileName) is False:
                            logger.debug("Renaming %s to %s",
                                         os.path.join(dataDirForDate, memberFileName),
                                         newFileName)
                            os.rename(os.path.join(dataDirForDate, memberFileName),
                                      newFileName)
                            uncompressSize = uncompressSize + memberZipInfo.file_size
                    except Exception as e:
                        logger.error("Error exptracting file %s from the downloaded zip archive: %s", memberFileName, e)
            # close the zip file
            zipDatafile.close()
            os.remove(zipFileName)  # delete zip file as its no longer required
        except Exception as e:
            logger.error("Error expanding archive file: %s", e)
        return(uncompressSize)

    def calculateFileNamePrefix(self, oldName, dataDirForDate, publishDateStr):
        """ calculate FileName Prefix """
        filename_part, file_extension = os.path.splitext(oldName)
        if oldName.startswith('An'):
            newFileNamePrefix = os.path.join(dataDirForDate,
                                             "equity_nse_announce_" + publishDateStr + file_extension)
        elif oldName.startswith('Bm'):
            newFileNamePrefix = os.path.join(dataDirForDate,
                                             "equity_nse_boardmeet_" + publishDateStr + file_extension)
        elif oldName.startswith('Bc'):
            newFileNamePrefix = os.path.join(dataDirForDate,
                                             "equity_nse_corp_action_" + publishDateStr + file_extension)
        elif oldName.startswith('Pd'):
            newFileNamePrefix = os.path.join(dataDirForDate,
                                             "equity_nse_" + publishDateStr + file_extension)
        elif oldName.startswith('corpbond'):
            newFileNamePrefix = os.path.join(dataDirForDate,
                                             "debt_nse_" + publishDateStr + file_extension)
        else:
            newFileNamePrefix = os.path.join(dataDirForDate, type(self).__name__ + "_" + oldName)
        return(newFileNamePrefix)

    def extractUniqueIDFromURL(self, URLToFetch):
        """ Get Unique ID From URL by extracting RegEx patterns matching any of urlMatchPatterns
        """
        # use today's date as default
        uniqueString = datetime.now().strftime('%d%m%y')
        date_obj = None
        if len(URLToFetch) > 6:
            for urlPattern in self.urlMatchPatterns:
                try:
                    result = urlPattern.search(URLToFetch)
                    uniqueString = result.group(2)
                    date_obj = datetime.strptime(uniqueString, '%d%m%y')
                    # if we did not encounter any error till this point, then this is the answer, so exit loop
                    break
                except Exception as e:
                    logger.debug("Error identifying unique ID of URL: %s , URL was: %s, Pattern: %s",
                                 e,
                                 URLToFetch.encode('ascii'),
                                 urlPattern)
        return((date_obj, uniqueString))

# # end of file ##
